#!/usr/bin/env Rscript
## ---------------------------
##
## Script name: process NASA GPW Data
##
## Purpose of script: process NASA GPW Data
## Author: Daniela Lührsen & Emily Ball
## Date Created: 2025-03-07
## Email: daniela.luhrsen@bsc.es
##
## ---------------------------

# Process NASA GPW Data
# https://cmr.earthdata.nasa.gov/search/concepts/C1597158029-SEDAC.html

# Load packages
library(raster)
library(dplyr)
library(purrr)
library(readr)

# Define bounding box filter early to reduce memory usage
bbox <- extent(-34, 45, 25, 75)

# File names and corresponding years
files <- list.files("data", full.names = TRUE)
files <- files[grepl("Global_\\d{4}", files)]
years <- c(2000, 2005, 2010, 2015, 2020)

# Preallocate list to store extracted data
extracted_data <- list()
coords <- NULL

for (i in seq_along(files)) {
  r <- raster(files[i])
  year <- years[i]

  r_crop <- crop(r, bbox)

  rm(r)
  # Convert to data frame with coordinates
  df <- as.data.frame(r_crop, xy = TRUE)
  rm(r_crop)

  # Only keep values within the bounding box
  df <- df %>% filter(x > -34 & x < 45 & y > 25 & y < 75)

  if (is.null(coords)) {
    coords <- df[, c("x", "y")]
  }

  # Store the values
  write_rds(df[[3]], paste0("data/gpw_", year, ".rds"))
  rm(df)
}

for (i in seq_along(files)) {
  year <- years[i]
  # Read the data back in
  df <- read_rds(paste0("data/gpw_", year, ".rds"))
  # Store in the list
  extracted_data[[i]] <- df
  rm(df)
}

# Combine into final data frame
final_df <- bind_cols(coords, setNames(as.data.frame(extracted_data), years))
rm(extracted_data)

final_df$population <- rowMeans(final_df[, 3:ncol(final_df)], na.rm = TRUE)
# Save the result
write_rds(final_df, "data/population.rds")