#!/usr/bin/env Rscript
## ---------------------------
##
## Script name: plot PET and precipitation
##
## Purpose of script: plot PET and precipitation, to be called by figure_02.R
##
## Author: Daniela Lührsen & Emily Ball
## Date Created: 2025-03-07
## Email: daniela.luhrsen@bsc.es
##
## ---------------------------

# month, region, accum variables imported from figure_02.R call

# load packages
packages <- c("sf", "zoo", "raster", "terra", "dplyr",
              "tidyr", "ggplot2", "stringr", "data.table")
lapply(packages, require, character.only = TRUE)

# load the data
prlr_masked <- unwrap(readRDS("../02_calculate_spei/data/temp/europe-masked_prlr.rds"))
pet_masked  <- unwrap(readRDS("../02_calculate_spei/data/temp/europe-masked_PET.rds"))

# Get mean across all pixels for each time step (monthly)
prlr_ts <- global(prlr_masked, fun = "mean", na.rm = TRUE)[, 1]
pet_ts <- global(pet_masked, fun = "mean", na.rm = TRUE)[, 1]

layer_names <- names(prlr_masked)
time_full <- str_extract(layer_names, "\\d{4}\\.\\d{2}.\\d{2}")
time_full <- str_replace(time_full, "\\.", "-")
time_full <- str_replace(time_full, "\\.", "-")

# get accumulation of 6 month
prlr_sum <- rollsum(prlr_ts, accum, align = "right")
pet_sum  <- rollsum(pet_ts, accum, align = "right")
time <- time_full[accum:length(time_full)]

## as in 2022 report, select month now to normalize timeseries of filtered data
idxs <- which(lubridate::month(time) == month)
prlr_sum <- prlr_sum[idxs]
pet_sum  <-  pet_sum[idxs]
time     <-     time[idxs]

normalize <- function(x) {
  (x - mean(x, na.rm = TRUE)) / sd(x, na.rm = TRUE)
}

prlr_norm <- normalize(prlr_sum)
pet_norm  <- normalize(pet_sum)
pet_norm_rev <- -1 * pet_norm

# Build data frame
df <- data.frame(
  Date = as.Date(paste0(time)),
  Precipitation = prlr_norm,
  PET = pet_norm_rev
)

# select month at this point if whole timeseries to be normalized
#df <- df[which(lubridate::month(df$Date) == 9), ]

# Reshape to long format
df_long <- melt(df, id.vars = "Date", variable.name = "Variable", value.name = "Value")
p_1 <- ggplot(df_long, aes(x = lubridate::year(Date), y = Value, color = Variable)) +
  geom_line(linewidth = 0.8) +
  geom_smooth(method = "loess", span = 0.2, se = FALSE, linetype = "dashed") +
  scale_color_manual(
    values = c("Precipitation" = "blue", "PET" = "red"),
    labels = c("Precipitation", "Potential Evapotranspiration")
  ) +
  scale_y_continuous(
    name = "Normalized Precipitation",
    sec.axis = sec_axis(~ -., name = "Normalized PET (reversed)")
  ) +
  geom_hline(yintercept = 0, linetype = "dotted") +
  labs(x = NULL, y = NULL) +
  theme_minimal(base_size = 14) +
  theme(
    legend.position = c(0.8,0.9),
    legend.background = element_rect(fill = alpha("white", 0.7), color = NA),
    legend.title = element_blank(),
    panel.grid.minor = element_blank(),
    plot.title = element_blank()
  )

final_plot <- p_1

# Save the plot as a PNG file
ggsave(paste0("figures/PET_pr_", region, "_spei_", accum, "_", month, ".png"),
       plot = final_plot, width = 12, height = 3, dpi = 300)

print(paste0("figures/PET_pr_", region, "_spei_", accum, "_", month, ".png"))

ggsave(paste0("figures/PET_pr_", region, "_spei_", accum, "_", month, ".pdf"),
       plot = final_plot, width = 12, height = 3, dpi = 300)

print(paste0("figures/PET_pr_", region, "_spei_", accum, "_", month, ".pdf"))
